# -*- coding: utf-8 -*-
import heimdall
import pytest
from ..stubs import *
from os.path import join, dirname, basename, realpath
from glob import glob


@pytest.mark.parametrize('db_tested', [
    ({'url': MENAGERIE_BASIC_XML, 'format': 'hera:xml'}),
    ({'url': MENAGERIE_BASIC_YAML, 'format': 'hera:yaml'}),
    ({'url': MENAGERIE_BASIC_JSON, 'format': 'hera:json'}),
    ({'url': MENAGERIE_BASIC_CSV, 'format': 'csv'}),
    ])
@pytest.mark.parametrize('db_expected', [
    ({'url': MENAGERIE_BASIC_XML, 'format': 'hera:xml'}),
    ({'url': MENAGERIE_BASIC_YAML, 'format': 'hera:yaml'}),
    ({'url': MENAGERIE_BASIC_JSON, 'format': 'hera:json'}),
    ({'url': MENAGERIE_BASIC_CSV, 'format': 'csv'}),
    ])
def test_get_database(db_tested, db_expected):
    # WHEN
    real = heimdall.getDatabase(**db_tested)
    expected = heimdall.getDatabase(**db_expected)
    # THEN
    properties = heimdall.getProperties(real)
    assert len(properties) == len(heimdall.getProperties(expected))
    entities = heimdall.getEntities(real)
    assert len(entities) == len(heimdall.getEntities(expected))
    for e in entities:
        eid = e.get('id')
        node = heimdall.getEntity(expected, lambda n: n.get('id') == eid)
        assert node is not None
        attributes = heimdall.getAttributes(e)
        assert len(attributes) == len(heimdall.getAttributes(node))
        for a in attributes:
            aid = a.get('id', None)
            if aid is not None:
                n = heimdall.getAttribute(node, lambda n: n.get('id') == aid)
                assert n is not None
    items = heimdall.getItems(real)
    assert len(items) == len(heimdall.getItems(expected))


@pytest.mark.parametrize(('content', 'format'), [
    ('<hera></hera>', 'hera:xml'),
    ('', 'hera:yaml'),
    ('{}', 'hera:yaml'),
    ('{}', 'hera:json'),
    ('', 'csv'),
    ])
def test_get_database_empty(tmp_path, content, format):
    # GIVEN
    path = tmp_path / 'empty.file'
    path.write_text(content)
    # WHEN
    tree = heimdall.getDatabase(url=str(path), format=format)
    # THEN
    assert len(heimdall.getProperties(tree)) == 0
    assert len(heimdall.getEntities(tree)) == 0
    assert len(heimdall.getItems(tree)) == 0


@pytest.mark.parametrize(('db', 'format', 'real_path', 'expected_path'), [
    # (menagerie_db, 'hera:xml', 'test.xml', MENAGERIE_BASIC_XML),
    (menagerie_db, 'hera:yaml', 'test.yaml', MENAGERIE_BASIC_YAML),
    (menagerie_db, 'hera:json', 'test.json', MENAGERIE_BASIC_JSON),
    ])
def test_serialize_to_file(tmp_path, db, format, real_path, expected_path):
    real_path = tmp_path / real_path
    # WHEN
    heimdall.serialize(db(), url=real_path, format=format)
    # THEN
    with open(real_path, 'r') as real:
        with open(expected_path, 'r') as expected:
            assert real.read().strip() == expected.read().strip()


@pytest.mark.parametrize(('db', 'format', 'expected_dir'), [
    (menagerie_db, 'csv', MENAGERIE_BASIC_CSV),
    ])
def test_serialize_to_dir(tmp_path, db, format, expected_dir):
    # WHEN
    heimdall.serialize(db(), url=tmp_path, format=format)
    # THEN
    real_dir = join(str(tmp_path), '*.csv')
    vs = dict()
    for expected_file in glob(expected_dir):
        key = basename(expected_file)
        vs[key] = [expected_file, ]
        for real_file in glob(real_dir):
            if real_file.endswith(key):
                vs[key].append(real_file)
    for key, paths in vs.items():
        expected_file = paths[0]
        real_file = paths[1]
        with open(real_file, 'r') as real:
            with open(expected_file, 'r') as expected:
                assert real.read().strip() == expected.read().strip()


@pytest.mark.parametrize(('content', 'format'), [
    # ('<hera></hera>', 'hera:xml'),
    ('{}', 'hera:yaml'),
    ('{}', 'hera:json'),
    ])
def test_serialize_empty(tmp_path, content, format):
    # GIVEN
    tree = like_rilly_empty_db()
    path = tmp_path / 'empty.file'
    # WHEN
    tree = heimdall.serialize(tree, url=str(path), format=format)
    # THEN
    assert path.read_text().strip() == content
